package bed

import (
	"fmt"
	"os"
	"runtime"

	"github.com/spf13/cobra"
	"golang.org/x/term"

	"b612.me/apps/b612/bed/cmdline"
	"b612.me/apps/b612/bed/editor"
	"b612.me/apps/b612/bed/tui"
	"b612.me/apps/b612/bed/window"
)

const (
	name     = "bed"
	version  = "0.2.8"
	revision = "HEAD"
)

var Cmd = &cobra.Command{
	Use:   "bed [文件路径]",
	Short: "基于 Go 开发的二进制文件编辑器",
	Long: `二进制文件编辑器 bed - 支持直接编辑二进制文件的命令行工具

支持功能：
  - 十六进制查看/编辑
  - 文件差异对比
  - 多窗口操作
  - 快速跳转地址`,
	Version:       fmt.Sprintf("%s (修订版本: %s/%s)", version, revision, runtime.Version()),
	SilenceUsage:  true,
	SilenceErrors: true,
	Args:          cobra.MaximumNArgs(1),
	RunE: func(cmd *cobra.Command, args []string) error {
		return runEditor(args)
	},
}

func init() {
	Cmd.SetVersionTemplate(`{{printf "%s 版本信息：" .Name}}{{.Version}}` + "\n")
	Cmd.Flags().BoolP("version", "v", false, "显示版本信息")
}

func runEditor(args []string) error {
	editor := editor.NewEditor(
		tui.NewTui(),
		window.NewManager(),
		cmdline.NewCmdline(),
	)

	if err := editor.Init(); err != nil {
		return fmt.Errorf("编辑器初始化失败: %w", err)
	}

	switch {
	case len(args) > 0 && args[0] != "-": // 处理文件参数
		if err := editor.Open(args[0]); err != nil {
			return fmt.Errorf("无法打开文件: %w", err)
		}
	case term.IsTerminal(int(os.Stdin.Fd())): // 交互模式
		if err := editor.OpenEmpty(); err != nil {
			return fmt.Errorf("创建空白文档失败: %w", err)
		}
	default: // 从标准输入读取
		if err := editor.Read(os.Stdin); err != nil {
			return fmt.Errorf("读取输入流失败: %w", err)
		}
	}

	defer editor.Close()
	return editor.Run()
}
