package gdu

import (
	"fmt"
	"os"
	"path/filepath"
	"regexp"
	"runtime"
	"strings"

	"github.com/gdamore/tcell/v2"
	"github.com/mattn/go-isatty"
	"github.com/rivo/tview"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"gopkg.in/yaml.v3"

	"b612.me/apps/b612/gdu/cmd/gdu/app"
	"b612.me/apps/b612/gdu/pkg/device"
)

var (
	af        *app.Flags
	configErr error
)

var Cmd = &cobra.Command{
	Use:   "gdu [directory_to_scan]",
	Short: "一款使用 Go 语言编写的快速磁盘空间分析工具。",
	Long: `一款使用 Go 语言编写的快速磁盘空间分析工具。

Gdu 主要针对 SSD 固态硬盘设计，能够充分利用并行处理优势。虽然也支持机械硬盘（HDD）使用，但性能提升效果不如前者显著。
`,
	Args:         cobra.MaximumNArgs(1),
	SilenceUsage: true,
	RunE:         runE,
}

func init() {
	af = &app.Flags{}
	flags := Cmd.Flags()
	flags.StringVar(&af.CfgFile, "config-file", "", "从配置文件读取（默认为 $HOME/.gdu.yaml）")
	flags.StringVarP(&af.LogFile, "log-file", "l", "/dev/null", "日志文件路径")
	flags.StringVarP(&af.OutputFile, "output-file", "o", "", "将所有信息导出为JSON文件")
	flags.StringVarP(&af.InputFile, "input-file", "f", "", "从JSON文件导入分析数据")
	flags.IntVarP(&af.MaxCores, "max-cores", "m", runtime.NumCPU(), fmt.Sprintf("设置Gdu使用的最大核心数。当前可用%d个核心", runtime.NumCPU()))
	flags.BoolVar(&af.SequentialScanning, "sequential", false, "使用顺序扫描（适用于机械硬盘HDD）")
	flags.BoolVarP(&af.ShowVersion, "version", "v", false, "打印版本信息")

	flags.StringSliceVarP(&af.IgnoreDirs, "ignore-dirs", "i", []string{"/proc", "/dev", "/sys", "/run"},
		"需要忽略的路径（逗号分隔），可为绝对路径或相对于当前目录的路径")
	flags.StringSliceVarP(&af.IgnoreDirPatterns, "ignore-dirs-pattern", "I", []string{},
		"需要忽略的路径模式（逗号分隔）")
	flags.StringVarP(&af.IgnoreFromFile, "ignore-from", "X", "",
		"从文件中读取需要忽略的路径模式")
	flags.BoolVarP(&af.NoHidden, "no-hidden", "H", false, "忽略隐藏目录（以点号开头的目录）")
	flags.BoolVarP(
		&af.FollowSymlinks, "follow-symlinks", "L", false,
		"跟踪文件的符号链接，显示链接指向文件的大小（不跟踪目录符号链接）",
	)
	flags.BoolVarP(
		&af.ShowAnnexedSize, "show-annexed-size", "A", false,
		"对git-annex文件显示表观大小（当文件未本地存储时，实际磁盘占用为零）",
	)
	flags.BoolVarP(&af.NoCross, "no-cross", "x", false, "不跨越文件系统边界")
	flags.BoolVarP(&af.ConstGC, "const-gc", "g", false, "启用恒定级别的内存垃圾回收（由GOGC参数控制）")
	flags.BoolVar(&af.Profiling, "enable-profiling", false, "启用性能分析数据收集（访问地址 http://localhost:6060/debug/pprof/）")

	flags.BoolVar(&af.UseStorage, "use-storage", false, "使用持久化键值存储分析数据（实验性功能）")
	flags.StringVar(&af.StoragePath, "storage-path", "/tmp/badger", "持久化键值存储目录路径")
	flags.BoolVarP(&af.ReadFromStorage, "read-from-storage", "r", false, "从持久化键值存储读取分析数据")

	flags.BoolVarP(&af.ShowDisks, "show-disks", "d", false, "显示所有已挂载磁盘")
	flags.BoolVarP(&af.ShowApparentSize, "show-apparent-size", "a", false, "显示表观大小")
	flags.BoolVarP(&af.ShowRelativeSize, "show-relative-size", "B", false, "显示相对大小")
	flags.BoolVarP(&af.NoColor, "no-color", "c", false, "禁用彩色输出")
	flags.BoolVarP(&af.ShowItemCount, "show-item-count", "C", false, "显示目录内项目数量")
	flags.BoolVarP(&af.ShowMTime, "show-mtime", "M", false, "显示目录内项目最新修改时间")
	flags.BoolVarP(&af.NonInteractive, "non-interactive", "n", false, "使用非交互模式")
	flags.BoolVarP(&af.NoProgress, "no-progress", "p", false, "非交互模式下不显示进度条")
	flags.BoolVarP(&af.NoUnicode, "no-unicode", "u", false, "禁用Unicode符号（用于大小进度条）")
	flags.BoolVarP(&af.Summarize, "summarize", "s", false, "非交互模式下仅显示统计总数")
	flags.IntVarP(&af.Top, "top", "t", 0, "非交互模式下仅显示前X个最大文件")
	flags.BoolVar(&af.UseSIPrefix, "si", false, "使用十进制SI单位（kB/MB/GB）而非二进制单位（KiB/MiB/GiB）")
	flags.BoolVar(&af.NoPrefix, "no-prefix", false, "非交互模式下显示原始数值（无单位前缀）")
	flags.BoolVar(&af.NoMouse, "no-mouse", false, "禁用鼠标支持")
	flags.BoolVar(&af.NoDelete, "no-delete", false, "禁止删除操作")
	flags.BoolVar(&af.WriteConfig, "write-config", false, "将当前配置写入文件（默认为 $HOME/.gdu.yaml）")

	initConfig()
	setDefaults()
}

func initConfig() {
	setConfigFilePath()
	data, err := os.ReadFile(af.CfgFile)
	if err != nil {
		configErr = err
		return // config file does not exist, return
	}

	configErr = yaml.Unmarshal(data, &af)
}

func setDefaults() {
	if af.Style.Footer.BackgroundColor == "" {
		af.Style.Footer.BackgroundColor = "#2479D0"
	}
	if af.Style.Footer.TextColor == "" {
		af.Style.Footer.TextColor = "#000000"
	}
	if af.Style.Footer.NumberColor == "" {
		af.Style.Footer.NumberColor = "#FFFFFF"
	}
	if af.Style.Header.BackgroundColor == "" {
		af.Style.Header.BackgroundColor = "#2479D0"
	}
	if af.Style.Header.TextColor == "" {
		af.Style.Header.TextColor = "#000000"
	}
	if af.Style.ResultRow.NumberColor == "" {
		af.Style.ResultRow.NumberColor = "#e67100"
	}
	if af.Style.ResultRow.DirectoryColor == "" {
		af.Style.ResultRow.DirectoryColor = "#3498db"
	}
}

func setConfigFilePath() {
	command := strings.Join(os.Args, " ")
	if strings.Contains(command, "--config-file") {
		re := regexp.MustCompile("--config-file[= ]([^ ]+)")
		parts := re.FindStringSubmatch(command)

		if len(parts) > 1 {
			af.CfgFile = parts[1]
			return
		}
	}
	setDefaultConfigFilePath()
}

func setDefaultConfigFilePath() {
	home, err := os.UserHomeDir()
	if err != nil {
		configErr = err
		return
	}

	path := filepath.Join(home, ".config", "gdu", "gdu.yaml")
	if _, err := os.Stat(path); err == nil {
		af.CfgFile = path
		return
	}

	af.CfgFile = filepath.Join(home, ".gdu.yaml")
}

func runE(command *cobra.Command, args []string) error {
	var (
		termApp *tview.Application
		screen  tcell.Screen
		err     error
	)

	if af.WriteConfig {
		data, err := yaml.Marshal(af)
		if err != nil {
			return fmt.Errorf("Error marshaling config file: %w", err)
		}
		if af.CfgFile == "" {
			setDefaultConfigFilePath()
		}
		err = os.WriteFile(af.CfgFile, data, 0o600)
		if err != nil {
			return fmt.Errorf("Error writing config file %s: %w", af.CfgFile, err)
		}
	}

	if runtime.GOOS == "windows" && af.LogFile == "/dev/null" {
		af.LogFile = "nul"
	}

	var f *os.File
	if af.LogFile == "-" {
		f = os.Stdout
	} else {
		f, err = os.OpenFile(af.LogFile, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, 0o600)
		if err != nil {
			return fmt.Errorf("opening log file: %w", err)
		}
		defer func() {
			cerr := f.Close()
			if cerr != nil {
				panic(cerr)
			}
		}()
	}
	log.SetOutput(f)

	if configErr != nil {
		log.Printf("Error reading config file: %s", configErr.Error())
	}

	istty := isatty.IsTerminal(os.Stdout.Fd())

	// we are not able to analyze disk usage on Windows and Plan9
	if runtime.GOOS == "windows" || runtime.GOOS == "plan9" {
		af.ShowApparentSize = true
	}

	if !af.ShowVersion && !af.NonInteractive && istty && af.OutputFile == "" {
		screen, err = tcell.NewScreen()
		if err != nil {
			return fmt.Errorf("Error creating screen: %w", err)
		}
		defer screen.Clear()
		defer screen.Fini()

		termApp = tview.NewApplication()
		termApp.SetScreen(screen)

		if !af.NoMouse {
			termApp.EnableMouse(true)
		}
	}

	a := app.App{
		Flags:       af,
		Args:        args,
		Istty:       istty,
		Writer:      os.Stdout,
		TermApp:     termApp,
		Screen:      screen,
		Getter:      device.Getter,
		PathChecker: os.Stat,
	}
	return a.Run()
}

func main() {
	if err := Cmd.Execute(); err != nil {
		os.Exit(1)
	}
}
